/*********** 
# Copyright (c) 2017, Xilinx, Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.
#
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT,INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY,OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
************/

/*****************************************************************************

    This example is intended to demonstrate custom data type usage in hardware
    function. In this example RGB color and HSV color structures are declared
    and used directly from hardware function.

*******************************************************************************/

#include <stdio.h>
#include <iostream>
#include <cstdio>
#include <stdlib.h>
#include <xil_io.h>
#include <unistd.h>
#include "xparameters.h"
#include "xmutex.h"

#include "rgb_to_hsv.h"
#include "sds_utils.h"

#define IMAGE_DIM 128

#include "mem_map.h"

#define MUTEX_DEVICE_ID		XPAR_MUTEX_0_IF_0_DEVICE_ID

/*
 * The Mutex number on which the example is run
 */
#define MUTEX_NUM 		0

/*
 * Mutex instance
 */
XMutex Mutex;

void start_Microblaze() {
    // file handles
    int valuefd, exportfd, directionfd;

    // The GPIO has to be exported to be able to see it
    // in sysfs

    exportfd = open("/sys/class/gpio/export", O_WRONLY);
    if (exportfd < 0)
    {
        printf("Cannot open GPIO to export\n");
        exit(1);
    }

    write(exportfd, "960", 4);
    close(exportfd);

    // Update the direction of the GPIO to be an output

    directionfd = open("/sys/class/gpio/gpio960/direction", O_RDWR);
    if (directionfd < 0)
    {
        printf("Cannot open GPIO direction it\n");
        exit(1);
    }

    write(directionfd, "out", 4);
    close(directionfd);

    // Get the GPIO value ready to be toggled

    valuefd = open("/sys/class/gpio/gpio960/value", O_RDWR);
    if (valuefd < 0)
    {
        printf("Cannot open GPIO value\n");
        exit(1);
    }

    // reset Microblaze
    write(valuefd,"0", 2);

    printf("Reset for 1 sec. ...\n");
    sleep(1);
    printf("Reset for 1 sec. done\n");

    // release MB reset
    write(valuefd,"1", 2);

    close(valuefd);

    //release exported gpio pin
    exportfd = open("/sys/class/gpio/unexport", O_WRONLY);
    if (exportfd < 0)
    {
        printf("Cannot open GPIO to unexport\n");
        exit(1);
    }

    write(exportfd, "960", 4);
    close(exportfd);
    return;
}

//Utility Function Declaration
void sw_RgbToHsv(RGBcolor *in, HSVcolor *out, int image_size);
void sw_HsvToRgb(HSVcolor *in, RGBcolor *out, int image_size);
int compareImages(unsigned int *in, unsigned int *out, int image_size);

int main(int argc, char* argv[])
{
	/*
	 * Init mutex and start microblaze
	 */
	XMutex_Config *ConfigPtr;
	XStatus Status;
	u32 TimeoutCount = 0;
	u32 val;

    u32 iiii=0;
    u32 value;

	int memfd;
	memfd = open("/dev/mem", O_RDWR | O_SYNC);
	if (memfd == -1) {
		printf("Can't open /dev/mem.\n");
		exit(0);
	}
	printf("/dev/mem opened.\n");

	/*
	 * Lookup configuration data in the device configuration table.
	 * Use this configuration info down below when initializing this
	 * driver instance.
	 */

	ConfigPtr = XMutex_LookupConfig(MUTEX_DEVICE_ID);
	if (ConfigPtr == (XMutex_Config *) NULL) {
		return XST_FAILURE;
	}

	/*
	 * Map Mutex memory to userspace
	 */
	mem_map_t mutex_map;

	if (map(memfd, (off_t) ConfigPtr->BaseAddress, &mutex_map)) {
		printf("Mutex mapping to userspace failed\r\n");
		exit(1);
	}

	ConfigPtr->BaseAddress = mutex_map.dev_base;

	/*
	 * Perform the rest of the initialization.
	 */
	Status = XMutex_CfgInitialize(&Mutex, ConfigPtr, ConfigPtr->BaseAddress);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	XMutex_SetUser(&Mutex, MUTEX_NUM, 0);
	XMutex_Lock(&Mutex, MUTEX_NUM);

	/*
	 * Map mb memory to userspace
	 */
	mem_map_t mb_map;

	if (maps(memfd, 0x28000000, &mb_map, 0x8000000)) {
		printf("Mb_map mapping to userspace failed\r\n");
		exit(1);
	}

	sds_mmap((void *) 0x28000000, 0x8000000, (void *) mb_map.dev_base);

		/*
	 * Write firmware for the EdkDSP at mapped 0x28000000
	 * It is up to 64 firmware programs, each 16kbytes = 4096 uint32 values
	 */

	printf("ARMCPU0: Write firmware ... \n\r");

	FILE *fr0, *fr1, *fr2, *fr3;

	/*
     * open input files to read
     */
    printf("ARMCPU0: Open  input file f0.dec ... ");
    if ( (fr0 = fopen("./f0.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    printf("ARMCPU0: Open  input file f1.dec ... ");
    if ( (fr1 = fopen("./f1.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

	printf("ARMCPU0: Open  input file f2.dec ... ");
    if ( (fr2 = fopen("./f2.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    printf("ARMCPU0: Open  input file f3.dec ... ");
    if ( (fr3 = fopen("./f3.dec", "r")) == NULL ) {
        printf("FAILED\n\r");
        return -1;
    }
    printf("OK\n\r");

    iiii = 0;
    value = 139266;
    while (fscanf(fr0,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(0*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(0*4096 + iiii), value);
	}

    iiii = 0;
    value = 139266;
    while (fscanf(fr1,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(1*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(1*4096 + iiii), value);
	}

	iiii = 0;
    value = 139266;
    while (fscanf(fr2,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(2*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(2*4096 + iiii), value);
	}
    
	iiii = 0;
    value = 139266;
    while (fscanf(fr3,"%d",&value)!=EOF){
    	Xil_Out32(mb_map.dev_base + 4*(3*4096 + iiii), value);
    	iiii++;
    }
    value=139266;
	for (; iiii < 4096; iiii++) {
		Xil_Out32(mb_map.dev_base + 4*(3*4096 + iiii), value);
	}

	/*
	 * close input files
	 */
	printf("ARMCPU0: Close input file f0.dec ... ");
	if ( fclose(fr0) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f1.dec ... ");
	if ( fclose(fr1) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f2.dec ... ");
	if ( fclose(fr2) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Close input file f3.dec ... ");
	if ( fclose(fr3) == EOF ) {
		printf("FAILED\n\r");
		return -1;
	}
	printf("OK\n\r");

	printf("ARMCPU0: Write firmware Done. \n\r");

	/*
	 * Initial waiting loop for Microblaze mapped at 0x28100000
	 */

#if (1)
	printf("ARMCPU0: place 0x28100000 at start of MB0 vectors\n\r");
	Xil_Out32(mb_map.dev_base + 0x00100000, 0xb8000000);
#endif

	/*
	 * Remove Microblaze reset via system control of an emio pin
	 */

	start_Microblaze();

	printf("ARMCPU0: MB0 reset removed, ARM waiting ...\n\r");

	/*
	 * Wait for response from Microblaze via user value at mutex 0
	 */
	do {
		XMutex_GetUser(&Mutex, MUTEX_NUM, &val);
	} while (val == 0);

	printf("ARMCPU0: MB0 indicates - running ...\n\r");

    // Size of synthetic image
    int image_size = IMAGE_DIM * IMAGE_DIM;

    // Allocate buffers using sds_alloc and malloc
    RGBcolor *rgbImage   = (RGBcolor *)(sds_alloc(sizeof(RGBcolor) * image_size));
    HSVcolor *hwHsvImage = (HSVcolor *)(sds_alloc(sizeof(HSVcolor) * image_size));
    HSVcolor *swHsvImage = (HSVcolor *)(malloc(sizeof(HSVcolor) * image_size));

    // Check for failed memory allocation
    if((swHsvImage == NULL) || (hwHsvImage == NULL) || (rgbImage == NULL)){
      std::cout << "TEST FAILED : Failed to allocate memory" << std::endl;
      return -1;
    }    

    // Initialize Test Input Data randomly
    for(int i = 0;i < image_size; i++){
        rgbImage[i].r = rand() % 256 ;
        rgbImage[i].g = rand() % 256 ;
        rgbImage[i].b = rand() % 256 ;
        rgbImage[i].pad = 0;
    }

    sds_utils::perf_counter hw_ctr, sw_ctr;

    hw_ctr.start();
    //Launch the Hardware Function
    rgb_to_hsv_accel(rgbImage, hwHsvImage, image_size);
    hw_ctr.stop();

    //Calculating sw based HSV Image
    sw_ctr.start();
    sw_RgbToHsv(rgbImage, swHsvImage, image_size);
    sw_ctr.stop();

    uint64_t sw_cycles = sw_ctr.avg_cpu_cycles();
    uint64_t hw_cycles = hw_ctr.avg_cpu_cycles();
    double speedup = (double) sw_cycles / (double) hw_cycles;

    std::cout << "Number of CPU cycles running application in software: "
        << sw_cycles << std::endl;
    std::cout << "Number of CPU cycles running application in hardware: "
        << hw_cycles << std::endl;
    std::cout << "Speed up: " << speedup << std::endl;
    std::cout << "Note: Speed up is meaningful for real hardware execution only, not for emulation." << std::endl;

    //Compare the results of the Hardware to the Sw Solution
    int match = compareImages((unsigned int *)swHsvImage, 
            (unsigned int *)hwHsvImage, image_size);

    // Release Memory 
    free(swHsvImage);
    sds_free(rgbImage);
    sds_free(hwHsvImage);

    if (match){
        std::cout << "TEST FAILED." << std::endl;
        return -1;
    }
    std::cout << "TEST PASSED." << std::endl;

	/*
	 *  unmap the memory before exiting
	 */
	if (munmap(mb_map.base, MAP_SIZE) == -1) {
		printf("Can't unmap memory from user space.\n");
		exit(0);
	}

	if (munmap(mutex_map.base, MAP_SIZE) == -1) {
		printf("Can't unmap memory from user space.\n");
		exit(0);
	}

	close(memfd);

    return 0;
}


// Utility Functions Definitions Start Here

// Convert RGB to HSV Format
void sw_RgbToHsv(RGBcolor* in, HSVcolor *out, int image_size)
{
    for(unsigned int i = 0 ; i < image_size; i++){
        RGBcolor rgb = in[i];
        HSVcolor hsv;
        
        unsigned char rgbMin, rgbMax;

        rgbMin = imin(rgb.r, (imin(rgb.g,rgb.b)));
        rgbMax = imax(rgb.r, (imax(rgb.g,rgb.b)));

        hsv.v = rgbMax;
        if (hsv.v == 0)
        {
            hsv.h = 0;
            hsv.s = 0;
            continue;
        }

        hsv.s = 255 * ((long)(rgbMax - rgbMin)) / hsv.v;
        if (hsv.s == 0)
        {
            hsv.h = 0;
            continue;
        }

        if (rgbMax == rgb.r)
            hsv.h = 0 + 43 * (rgb.g - rgb.b) / (rgbMax - rgbMin);
        else if (rgbMax == rgb.g)
            hsv.h = 85 + 43 * (rgb.b - rgb.r) / (rgbMax - rgbMin);
        else
            hsv.h = 171 + 43 * (rgb.r - rgb.g) / (rgbMax - rgbMin);

        hsv.pad = rgb.pad;
        out[i] = hsv;
    }
}

// Convert RGB to HSV Format
void sw_HsvToRgb(HSVcolor *in, RGBcolor *out, int image_size)
{
    for(int i = 0; i < image_size; i++)
    {
        RGBcolor rgb;
        HSVcolor hsv = in[i];
        unsigned char region, p, q, t;
        unsigned int h, s, v, remainder;

        if (hsv.s == 0)
        {
            rgb.r = hsv.v;
            rgb.g = hsv.v;
            rgb.b = hsv.v;
            continue;
        }

        // converting to 16 bit to prevent overflow
        h = hsv.h;
        s = hsv.s;
        v = hsv.v;

        region = h / 43;
        remainder = (h - (region * 43)) * 6;

        p = (v * (255 - s)) >> 8;
        q = (v * (255 - ((s * remainder) >> 8))) >> 8;
        t = (v * (255 - ((s * (255 - remainder)) >> 8))) >> 8;

        switch (region)
        {
            case 0:
                rgb.r = v;
                rgb.g = t;
                rgb.b = p;
                break;
            case 1:
                rgb.r = q;
                rgb.g = v;
                rgb.b = p;
                break;
            case 2:
                rgb.r = p;
                rgb.g = v;
                rgb.b = t;
                break;
            case 3:
                rgb.r = p;
                rgb.g = q;
                rgb.b = v;
                break;
            case 4:
                rgb.r = t;
                rgb.g = p;
                rgb.b = v;
                break;
            default:
                rgb.r = v;
                rgb.g = p;
                rgb.b = q;
                break;
        }
        rgb.pad = hsv.pad;
        out[i] = rgb;
    }
}

int compareImages(unsigned int *img1, unsigned int *img2, int image_size)
{
    for (int i = 0, cnt = 0 ; i < image_size ; i++)
    {
        unsigned int in  = img1[i];
        unsigned int out = img2[i];
        if (in != out ){
            cnt++;
            std::cout << "ERROR: Pixel=" << i << " mismatch Expected="
               <<  std::hex << in << " and Got=" << out << std::endl;
            return -1;
        }
    }
    return 0;
}
